from os import path as os_path, pardir as os_pardir, name as os_name
from sys import version_info


## Python version 
PYTHON_VERSION = version_info.major

## Maximum number or pages to search
SEARCH_ENGINE_RESULTS_PAGES = 20

## HTTP request timeout 
TIMEOUT = 10

## Default User-Agent string 
USER_AGENT = 'search_engines/0.5 Repo: https://github.com/tasos-py/Search-Engines-Scraper'

## Fake User-Agent string - Google desn't like the default user-agent
FAKE_USER_AGENT = 'Mozilla/5.0 (Windows NT 6.1; rv:84.0) Gecko/20100101 Firefox/84.0'

## Proxy server 
PROXY = None

## TOR proxy server 
TOR = 'socks5h://127.0.0.1:9050'

_base_dir = os_path.abspath(os_path.dirname(os_path.abspath(__file__)))

## Path to output files 
OUTPUT_DIR = os_path.join(_base_dir, 'search_results') + os_path.sep




from bs4 import BeautifulSoup

from ..engine import SearchEngine
from ..config import PROXY, TIMEOUT, FAKE_USER_AGENT
from .. import output as out


class Startpage(SearchEngine):
    '''Searches startpage.com'''
    def __init__(self, proxy=PROXY, timeout=TIMEOUT): 
        super(Startpage, self).__init__(proxy, timeout)
        self._base_url = 'https://www.startpage.com'
        self.set_headers({'User-Agent':FAKE_USER_AGENT})
    
    def _selectors(self, element):
        '''Returns the appropriate CSS selector.'''
        selectors = {
            'url': 'a.w-gl__result-url', 
            'title': 'a.w-gl__result-title h3', 
            'text': 'p.w-gl__description', 
            'links': 'section.w-gl div.w-gl__result', 
            'next': {'form':'form.pagination__form', 'text':'Next'},
            'search_form': 'form#search input[name]',
            'blocked_form': 'form#blocked_feedback_form'
        }
        return selectors[element]
    
    def _first_page(self):
        '''Returns the initial page and query.'''
        response = self._get_page(self._base_url)
        tags = BeautifulSoup(response.html, "html.parser")
        selector = self._selectors('search_form')

        data = {
            i['name']: i.get('value', '') 
            for i in tags.select(selector)
        }
        data['query'] = self._query
        url = self._base_url + '/sp/search'
        return {'url':url, 'data':data}
    
    def _next_page(self, tags):
        '''Returns the next page URL and post data (if any)'''
        selector = self._selectors('next')
        forms = [
            form 
            for form in tags.select(selector['form']) 
            if form.get_text(strip=True) == selector['text']
        ]
        url, data = None, None
        if forms:
            url = self._base_url + forms[0]['action']
            data = {
                i['name']:i.get('value', '') 
                for i in forms[0].select('input')
            }
        return {'url':url, 'data':data}
    
    def _is_ok(self, response):
        '''Checks if the HTTP response is 200 OK.'''
        soup = BeautifulSoup(response.html, 'html.parser')
        selector = self._selectors('blocked_form')
        is_blocked = soup.select_one(selector)
        
        self.is_banned = response.http in [403, 429, 503] or is_blocked
        
        if response.http == 200 and not is_blocked:
            return True
        msg = 'Banned' if is_blocked else ('HTTP ' + str(response.http)) if response.http else response.html
        out.console(msg, level=out.Level.error)
        return False



import requests
from requests.auth import HTTPProxyAuth, HTTPBasicAuth
from collections import namedtuple

from .config import TIMEOUT, PROXY, USER_AGENT
from . import utils as utl


class HttpClient(object):
    '''Performs HTTP requests. A `requests` wrapper, essentialy'''
    def __init__(self, timeout=TIMEOUT, proxy=PROXY, username: str or None = None, password: str or None = None):
        self.session = requests.session()
        self.session.proxies = self._set_proxy(proxy)
        self.session.auth = self._set_auth(username, password)
        self.session.headers['User-Agent'] = USER_AGENT
        self.session.headers['Accept-Language'] = 'en-GB,en;q=0.5'

        self.timeout = timeout
        self.response = namedtuple('response', ['http', 'html'])

    def get(self, page):
        '''Submits a HTTP GET request.'''
        #print("Getting...")
        page = self._quote(page) #
        try:
            req = self.session.get(page, timeout=self.timeout)
            #print(req.status_code) #, req.raw._connection.sock.getsockname())
            self.session.headers['Referer'] = page
        except requests.exceptions.RequestException as e:
            #print(str(e))
            return self.response(http=0, html=e.__doc__)
        
        return self.response(http=req.status_code, html=req.text)
    
    def post(self, page, data):
        '''Submits a HTTP POST request.'''
        page = self._quote(page)
        try:
            req = self.session.post(page, data, timeout=self.timeout)
            self.session.headers['Referer'] = page
        except requests.exceptions.RequestException as e:
            return self.response(http=0, html=e.__doc__)
        return self.response(http=req.status_code, html=req.text)
    
    def _quote(self, url):
        '''URL-encodes URLs.'''
        if utl.decode_bytes(utl.unquote_url(url)) == utl.decode_bytes(url):
            url = utl.quote_url(url)
        return url
    
    def _set_proxy(self, proxy):
        '''Returns HTTP or SOCKS proxies dictionary.'''
        #if proxy:
        #    if not utl.is_url(proxy):
        #        raise ValueError('Invalid proxy format!')
        if proxy:
            proxy = {'http': f"http://{proxy}"} #, 'https': f"https://{proxy}"}
            return proxy
        return None

    def _set_auth(self, username: str, password: str): 
        '''Returns HTTP Proxy Auth'''
        #print(username, password)
        if username and password: 
            return HTTPProxyAuth(username, password)
        return None


from ..engine import SearchEngine
from ..config import PROXY, TIMEOUT


class Brave(SearchEngine):
    '''Searches brave.com'''
    def __init__(self, proxy=PROXY, timeout=TIMEOUT, username: str or None = None, password: str or None = None, language: str or None = None):
        super(Brave, self).__init__(proxy, timeout, username, password, language)
        self._base_url = 'https://search.brave.com'
    
    def _selectors(self, element):
        '''Returns the appropriate CSS selector.'''
        selectors = {
            'url': 'a.result-header[href]', 
            'title': 'a.result-header[href] span.snippet-title', 
            'text': 'div.snippet-content', 
            'links': 'div#results div[data-loc="main"]', 
            'next': {'tag':'div#pagination a[href]', 'text':'Next', 'skip':'disabled'}
        }
        return selectors[element]
    
    def _first_page(self):
        '''Returns the initial page and query.'''
        url_str = u'{}/search?q={}&source=web'
        url = url_str.format(self._base_url, self._query)
        return {'url':url, 'data':None}
    
    def _next_page(self, tags):
        '''Returns the next page URL and post data (if any)'''
        selector = self._selectors('next')
        next_page = [
            tag for tag in tags.select(selector['tag']) 
            if tag.get_text().strip() == selector['text'] and selector['skip'] not in tag['class']
        ]
        url = None
        if next_page:
            url = self._base_url + next_page[0]['href']
        return {'url':url, 'data':None}




from ..engine import SearchEngine
from ..config import PROXY, TIMEOUT, FAKE_USER_AGENT


class Mojeek(SearchEngine):
    '''Searches mojeek.com'''
    def __init__(self, proxy=PROXY, timeout=TIMEOUT):
        super(Mojeek, self).__init__(proxy, timeout)
        self._base_url = 'https://www.mojeek.com'
        self.set_headers({'User-Agent':FAKE_USER_AGENT})
    
    def _selectors(self, element):
        '''Returns the appropriate CSS selector.'''
        selectors = {
            'url': 'a.ob[href]', 
            'title': 'a.ob[href]', 
            'text': 'p.s', 
            'links': 'ul.results-standard > li', 
            'next': {'href':'div.pagination li a[href]', 'text':'Next'}
        }
        return selectors[element]
    
    def _first_page(self):
        '''Returns the initial page and query.'''
        url = u'{}/search?q={}'.format(self._base_url, self._query)
        return {'url':url, 'data':None}
    
    def _next_page(self, tags):
        '''Returns the next page URL and post data (if any)'''
        selector = self._selectors('next')
        next_page = [
            i['href'] for i in tags.select(selector['href']) 
            if i.text == selector['text']
        ]
        url = (self._base_url + next_page[0]) if next_page else None
        return {'url':url, 'data':None}




from .aol import Aol
from .ask import Ask
from .bing import Bing
from .dogpile import Dogpile
from .duckduckgo import Duckduckgo
from .google import Google
from .mojeek import Mojeek
from .startpage import Startpage
from .torch import Torch
from .yahoo import Yahoo
from .qwant import Qwant
from .brave import Brave


search_engines_dict = { 
    'google': Google, 
    'bing': Bing, 
    'yahoo': Yahoo, 
    'aol': Aol, 
    'duckduckgo': Duckduckgo, 
    'startpage': Startpage, 
    'dogpile': Dogpile, 
    'ask': Ask, 
    'mojeek': Mojeek, 
    'qwant': Qwant, 
    'brave': Brave,
    'torch': Torch 
}
#


from ..engine import SearchEngine
from ..config import TOR, TIMEOUT
from .. import output as out


class Torch(SearchEngine):
    '''Uses torch search engine. Requires TOR proxy.'''
    def __init__(self, proxy=TOR, timeout=TIMEOUT):
        super(Torch, self).__init__(proxy, timeout)
        self._base_url = u'http://torchdeedp3i2jigzjdmfpn5ttjhthh5wbmda2rr3jvqjg5p77c54dqd.onion'
        if not proxy:
            out.console('Torch requires TOR proxy!', level=out.Level.warning)
        self._current_page = 1
    
    def _selectors(self, element):
        '''Returns the appropriate CSS selector.'''
        selectors = {
            'url': 'h5 a[href]', 
            'title': 'h5 a[href]', 
            'text': 'p', 
            'links': 'div.result.mb-3', 
            'next': 'ul.pagination a.page-link'
        }
        return selectors[element]
    
    def _first_page(self):
        '''Returns the initial page and query.'''
        url_str = u'{}/search?query={}&action=search'
        url = url_str.format(self._base_url, self._query)
        return {'url':url, 'data':None}
    
    def _next_page(self, tags):
        '''Returns the next page URL and post data (if any)'''
        self._current_page += 1
        url_str = u'{}/search?query={}&page={}'
        url = url_str.format(self._base_url, self._query, self._current_page)
        return {'url':url, 'data':None}



from ..engine import SearchEngine
from ..config import PROXY, TIMEOUT, FAKE_USER_AGENT


class Bing(SearchEngine):
    '''Searches bing.com'''
    def __init__(self, proxy=PROXY, timeout=TIMEOUT, username: str or None = None, password: str or None = None, language: str or None = None):
        super(Bing, self).__init__(proxy, timeout, username, password, language)
        self._base_url = u'https://www.bing.com'
        self.set_headers({'User-Agent':FAKE_USER_AGENT})
        self.__language_dict = {
            "EN": "en",
            "ES": "es"
        }

    def _selectors(self, element):
        '''Returns the appropriate CSS selector.'''
        selectors = {
            'url': 'a[href]', 
            'title': 'h2', 
            'text': 'p', 
            'links': 'ol#b_results > li.b_algo', 
            'next': 'div#b_content nav[role="navigation"] a.sb_pagN'
        }
        return selectors[element]
    
    def _first_page(self):
        '''Returns the initial page and query.'''
        self._get_page(self._base_url)
        language = self.__language_dict[self._language] if self._language in self.__language_dict else "en"
        url = u'{}/search?q={}&search=&form=QBLH&setlang={}'.format(self._base_url, self._query, language)
        return {'url':url, 'data':None}

    def _next_page(self, tags):
        '''Returns the next page URL and post data (if any)'''
        selector = self._selectors('next')
        next_page = self._get_tag_item(tags.select_one(selector), 'href')
        url = None
        if next_page:
            url = (self._base_url + next_page) 
        return {'url':url, 'data':None}
